@@ -116,13 +116,13 @@ get_type_blob (GTypelib *typelib,
       return FALSE;
     }
 
-  if (simple->reserved == 0 && simple->reserved2 == 0)
+  if (simple->flags.reserved == 0 && simple->flags.reserved2 == 0)
     {
       g_set_error (error,
 		   G_TYPELIB_ERROR,
 		   G_TYPELIB_ERROR_INVALID,
 		   "Expected non-basic type but got %d",
-		   simple->tag);
+		   simple->flags.tag);
       return FALSE;
     }
 
@@ -161,31 +161,32 @@ g_typelib_check_sanity (void)
    * Everything else in the code however should be using sizeof().
    */
   
-  CHECK_SIZE (Header, 108);
+  CHECK_SIZE (Header, 112);
   CHECK_SIZE (DirEntry, 12);
   CHECK_SIZE (SimpleTypeBlob, 4);
-  CHECK_SIZE (ArgBlob, 12);
+  CHECK_SIZE (ArgBlob, 16);
   CHECK_SIZE (SignatureBlob, 8);
   CHECK_SIZE (CommonBlob, 8);
   CHECK_SIZE (FunctionBlob, 20);
+  CHECK_SIZE (CallbackBlob, 12);
   CHECK_SIZE (InterfaceTypeBlob, 4);
   CHECK_SIZE (ArrayTypeBlob, 8);
   CHECK_SIZE (ParamTypeBlob, 4);
   CHECK_SIZE (ErrorTypeBlob, 4);
   CHECK_SIZE (ErrorDomainBlob, 16);
   CHECK_SIZE (ValueBlob, 12);
-  CHECK_SIZE (FieldBlob, 12);
+  CHECK_SIZE (FieldBlob, 16);
   CHECK_SIZE (RegisteredTypeBlob, 16);
-  CHECK_SIZE (StructBlob, 24);
-  CHECK_SIZE (EnumBlob, 20);
-  CHECK_SIZE (PropertyBlob, 12);
-  CHECK_SIZE (SignalBlob, 12);
-  CHECK_SIZE (VFuncBlob, 16);
-  CHECK_SIZE (ObjectBlob, 32);
-  CHECK_SIZE (InterfaceBlob, 28);
-  CHECK_SIZE (ConstantBlob, 20);
-  CHECK_SIZE (AnnotationBlob, 12);
-  CHECK_SIZE (UnionBlob, 32);
+  CHECK_SIZE (StructBlob, 32);
+  CHECK_SIZE (EnumBlob, 24);
+  CHECK_SIZE (PropertyBlob, 16);
+  CHECK_SIZE (SignalBlob, 16);
+  CHECK_SIZE (VFuncBlob, 20);
+  CHECK_SIZE (ObjectBlob, 44);
+  CHECK_SIZE (InterfaceBlob, 40);
+  CHECK_SIZE (ConstantBlob, 24);
+  CHECK_SIZE (AttributeBlob, 12);
+  CHECK_SIZE (UnionBlob, 40);
 #undef CHECK_SIZE
 
   g_assert (size_check_ok);
@@ -286,7 +287,7 @@ validate_header (ValidateContext  *ctx,
       
     }
 
-  if (header->major_version != 1 || header->minor_version != 0)
+  if (header->major_version != 2 || header->minor_version != 0)
     {
       g_set_error (error,
 		   G_TYPELIB_ERROR,
@@ -313,25 +314,33 @@ validate_header (ValidateContext  *ctx,
 		   "Typelib size mismatch");
       return FALSE; 
     }
+  
+  /* This is a sanity check for a specific typelib; it
+   * prevents us from loading an incompatible typelib.
+   * 
+   * The hardcoded checks in g_typelib_check_sanity to
+   * protect against inadvertent or buggy changes to the typelib format
+   * itself.
+   */
 
-  if (header->entry_blob_size != 12 ||
-      header->function_blob_size != 20 ||
-      header->callback_blob_size != 12 ||
-      header->signal_blob_size != 12 ||
-      header->vfunc_blob_size != 16 ||
-      header->arg_blob_size != 12 ||
-      header->property_blob_size != 12 ||
-      header->field_blob_size != 12 ||
-      header->value_blob_size != 12 ||
-      header->constant_blob_size != 20 ||
-      header->error_domain_blob_size != 16 ||
-      header->annotation_blob_size != 12 ||
-      header->signature_blob_size != 8 ||
-      header->enum_blob_size != 20 ||
-      header->struct_blob_size != 24 ||
-      header->object_blob_size != 32 ||
-      header->interface_blob_size != 28 ||
-      header->union_blob_size != 32)
+  if (header->entry_blob_size != sizeof (DirEntry) ||
+      header->function_blob_size != sizeof (FunctionBlob) ||
+      header->callback_blob_size != sizeof (CallbackBlob) ||
+      header->signal_blob_size != sizeof (SignalBlob) ||
+      header->vfunc_blob_size != sizeof (VFuncBlob) ||
+      header->arg_blob_size != sizeof (ArgBlob) ||
+      header->property_blob_size != sizeof (PropertyBlob) ||
+      header->field_blob_size != sizeof (FieldBlob) ||
+      header->value_blob_size != sizeof (ValueBlob) ||
+      header->constant_blob_size != sizeof (ConstantBlob) ||
+      header->error_domain_blob_size != sizeof (ErrorDomainBlob) ||
+      header->attribute_blob_size != sizeof (AttributeBlob) ||
+      header->signature_blob_size != sizeof (SignatureBlob) ||
+      header->enum_blob_size != sizeof (EnumBlob) ||
+      header->struct_blob_size != sizeof (StructBlob) ||
+      header->object_blob_size != sizeof(ObjectBlob) ||
+      header->interface_blob_size != sizeof (InterfaceBlob) ||
+      header->union_blob_size != sizeof (UnionBlob))
     {
       g_set_error (error,
 		   G_TYPELIB_ERROR,
@@ -349,21 +358,21 @@ validate_header (ValidateContext  *ctx,
       return FALSE; 
     }
 
-  if (!is_aligned (header->annotations))
+  if (!is_aligned (header->attributes))
     {
       g_set_error (error,
 		   G_TYPELIB_ERROR,
 		   G_TYPELIB_ERROR_INVALID_HEADER,
-		   "Misaligned annotations");
+		   "Misaligned attributes");
       return FALSE; 
     }
 
-  if (header->annotations == 0 && header->n_annotations > 0)
+  if (header->attributes == 0 && header->n_attributes > 0)
     {
       g_set_error (error,
 		   G_TYPELIB_ERROR,
 		   G_TYPELIB_ERROR_INVALID_HEADER,
-		   "Wrong number of annotations");
+		   "Wrong number of attributes");
       return FALSE; 
     }
 
@@ -538,10 +547,10 @@ validate_type_blob (GTypelib     *typelib,
   
   simple = (SimpleTypeBlob *)&typelib->data[offset];
 
-  if (simple->reserved == 0 && 
-      simple->reserved2 == 0)
+  if (simple->flags.reserved == 0 && 
+      simple->flags.reserved2 == 0)
     {
-      if (simple->tag >= GI_TYPE_TAG_ARRAY)
+      if (simple->flags.tag >= GI_TYPE_TAG_ARRAY)
 	{
 	  g_set_error (error,
 		       G_TYPELIB_ERROR,
@@ -550,13 +559,13 @@ validate_type_blob (GTypelib     *typelib,
 	  return FALSE;
 	}
       
-      if (simple->tag >= GI_TYPE_TAG_UTF8 &&
-	  !simple->pointer)
+      if (simple->flags.tag >= GI_TYPE_TAG_UTF8 &&
+	  !simple->flags.pointer)
 	{
 	  g_set_error (error,
 		       G_TYPELIB_ERROR,
 		       G_TYPELIB_ERROR_INVALID_BLOB,
-		       "Pointer type exected for tag %d", simple->tag);
+		       "Pointer type exected for tag %d", simple->flags.tag);
 	  return FALSE;	  
 	}
 
@@ -895,6 +904,8 @@ validate_constant_blob (GTypelib     *typelib,
     4, /* UINT32 */
     8, /* INT64 */
     8, /* UINT64 */
+    sizeof (gshort),
+    sizeof (gushort),
     sizeof (gint),
     sizeof (guint),
     sizeof (glong),
@@ -954,9 +965,9 @@ validate_constant_blob (GTypelib     *typelib,
     }
   
   type = (SimpleTypeBlob *)&typelib->data[offset + G_STRUCT_OFFSET (ConstantBlob, type)];
-  if (type->reserved == 0 && type->reserved2 == 0)
+  if (type->flags.reserved == 0 && type->flags.reserved2 == 0)
     {
-      if (type->tag == 0)
+      if (type->flags.tag == 0)
 	{
 	  g_set_error (error,
 		       G_TYPELIB_ERROR,
@@ -965,8 +976,8 @@ validate_constant_blob (GTypelib     *typelib,
 	  return FALSE;
 	}
 
-      if (value_size[type->tag] != 0 &&
-	  blob->size != value_size[type->tag])
+      if (value_size[type->flags.tag] != 0 &&
+	  blob->size != value_size[type->flags.tag])
 	{
 	  g_set_error (error,
 		       G_TYPELIB_ERROR,
@@ -1435,7 +1446,9 @@ validate_object_blob (ValidateContext *ctx,
     {
       DirEntry *entry;
 
-      entry = g_typelib_get_dir_entry (typelib, blob->parent);
+      entry = get_dir_entry_checked (typelib, blob->parent, error);
+      if (!entry)
+        return FALSE;
       if (entry->blob_type != BLOB_TYPE_OBJECT &&
 	  (entry->local || entry->blob_type != 0))
 	{
@@ -1447,6 +1460,23 @@ validate_object_blob (ValidateContext *ctx,
 	}
     }
   
+  if (blob->gtype_struct != 0)
+    {
+      DirEntry *entry;
+
+      entry = get_dir_entry_checked (typelib, blob->gtype_struct, error);
+      if (!entry)
+        return FALSE;
+      if (entry->blob_type != BLOB_TYPE_STRUCT && entry->local)
+        {
+          g_set_error (error,
+                       G_TYPELIB_ERROR,
+                       G_TYPELIB_ERROR_INVALID_BLOB,
+                       "Class struct invalid type or not local");
+          return FALSE; 
+        }
+    }  
+  
   if (typelib->len < offset + sizeof (ObjectBlob) + 
             (blob->n_interfaces + blob->n_interfaces % 2) * 2 +
             blob->n_fields * sizeof (FieldBlob) +
@@ -1481,7 +1511,9 @@ validate_object_blob (ValidateContext *ctx,
 	  return FALSE; 
 	}
       
-      entry = g_typelib_get_dir_entry (typelib, iface);
+      entry = get_dir_entry_checked (typelib, iface, error);
+      if (!entry)
+        return FALSE;
 
       if (entry->blob_type != BLOB_TYPE_INTERFACE &&
 	  (entry->local || entry->blob_type != 0))
@@ -1830,13 +1862,13 @@ validate_directory (ValidateContext   *ctx,
 }
 
 static gboolean
-validate_annotations (ValidateContext *ctx, 
-		      GError       **error)
+validate_attributes (ValidateContext *ctx,
+		     GError       **error)
 {
   GTypelib *typelib = ctx->typelib;
   Header *header = (Header *)typelib->data;
 
-  if (header->size < header->annotations + header->n_annotations * sizeof (AnnotationBlob))
+  if (header->size < header->attributes + header->n_attributes * sizeof (AttributeBlob))
     {
       g_set_error (error,
 		   G_TYPELIB_ERROR,
@@ -1896,9 +1928,9 @@ g_typelib_validate (GTypelib     *typelib,
       return FALSE;
     }
 
-  if (!validate_annotations (&ctx, error))
+  if (!validate_attributes (&ctx, error))
     {
-      prefix_with_context (error, "annotations", &ctx);
+      prefix_with_context (error, "attributes", &ctx);
       return FALSE;
     }
 
@@ -1914,84 +1946,99 @@ g_typelib_error_quark (void)
   return quark;
 }
 
-static inline void
-_g_typelib_init (GTypelib *typelib)
+static void
+_g_typelib_do_dlopen (GTypelib *typelib)
 {
   Header *header;
-
+  const char *shlib_str;
+  
   header = (Header *) typelib->data;
+  /* note that NULL shlib means to open the main app, which is allowed */
   if (header->shared_library)
+    shlib_str = g_typelib_get_string (typelib, header->shared_library);
+  else
+    shlib_str = NULL;
+
+  if (shlib_str != NULL && shlib_str[0] != '\0')
     {
-      const gchar *shlib_str;
-      GModule *app_module = NULL;
+      gchar **shlibs;
+      gint i;
 
-      shlib_str = g_typelib_get_string (typelib, header->shared_library);
-      /* note that NULL shlib means to open the main app, which is allowed */
+      /* shared-library is a comma-separated list of libraries */
+      shlibs = g_strsplit (shlib_str, ",", 0);
 
-      if (shlib_str != NULL)
+       /* We load all passed libs unconditionally as if the same library is loaded
+        * again with dlopen(), the same file handle will be returned. See bug:
+        * http://bugzilla.gnome.org/show_bug.cgi?id=555294
+        */
+      for (i = 0; shlibs[i]; i++)
         {
-          gchar **shlibs;
-          gint i;
+          GModule *module;
 
-          /* shared-library is a comma-separated list of libraries */
-          shlibs = g_strsplit (shlib_str, ",", 0);
-
-          /* We load all passed libs unconditionally as if the same library is loaded
-           * again with dlopen(), the same file handle will be returned. See bug:
-           * http://bugzilla.gnome.org/show_bug.cgi?id=555294
+          /* Glade's autoconnect feature and OpenGL's extension mechanism
+           * as used by Clutter rely on dlopen(NULL) to work as a means of
+           * accessing the app's symbols. This keeps us from using
+           * G_MODULE_BIND_LOCAL. BIND_LOCAL may have other issues as well;
+           * in general libraries are not expecting multiple copies of
+           * themselves and are not expecting to be unloaded. So we just
+           * load modules globally for now.
            */
-          for (i = 0; shlibs[i]; i++)
+
+          module = g_module_open (shlibs[i], G_MODULE_BIND_LAZY);
+
+          if (module == NULL)
             {
-              GModule *module;
+              GString *shlib_full = g_string_new (shlibs[i]);
 
-              /* Glade's autoconnect feature and OpenGL's extension mechanism
-               * as used by Clutter rely on dlopen(NULL) to work as a means of
-               * accessing the app's symbols. This keeps us from using
-               * G_MODULE_BIND_LOCAL. BIND_LOCAL may have other issues as well;
-               * in general libraries are not expecting multiple copies of
-               * themselves and are not expecting to be unloaded. So we just
-               * load modules globally for now.
-               */
-
-	      module = g_module_open (shlibs[i], G_MODULE_BIND_LAZY);
-
-	      if (module == NULL)
-	        {
-	          GString *shlib_full = g_string_new (shlibs[i]);
-
-	          /* Prefix with "lib", try both .la and .so */
-	          if (!g_str_has_prefix (shlib_full->str, "lib"))
-	            g_string_prepend (shlib_full, "lib");
-	          g_string_append (shlib_full, ".la");
-	          module = g_module_open (shlib_full->str, G_MODULE_BIND_LAZY);
-	          if (module == NULL)
-	            g_string_overwrite (shlib_full, strlen (shlib_full->str)-2, SHLIB_SUFFIX);
-	          module = g_module_open (shlib_full->str, G_MODULE_BIND_LAZY);
-
-	          g_string_free (shlib_full, TRUE);
-	        }
-
-	      if (module == NULL)
+              /* Prefix with "lib", try both .la and .so */
+              if (!g_str_has_prefix (shlib_full->str, "lib"))
+                g_string_prepend (shlib_full, "lib");
+              g_string_append (shlib_full, ".la");
+              module = g_module_open (shlib_full->str, G_MODULE_BIND_LAZY);
+              if (module == NULL)
                 {
-                  g_warning ("Failed to load shared library '%s' referenced by the typelib: %s",
-                             shlibs[i], g_module_error ());
+                  g_string_overwrite (shlib_full, strlen (shlib_full->str)-2, SHLIB_SUFFIX);
+                  module = g_module_open (shlib_full->str, G_MODULE_BIND_LAZY);
                 }
-              else
-                {
-                  typelib->modules = g_list_append (typelib->modules, module);
-                }
-          }
 
-          g_strfreev (shlibs);
-        }
+              g_string_free (shlib_full, TRUE);
+            }
 
-        /* we should make sure the app_module in the end of list so that
-         * it's last symbol source when loading any symbols from modules.
-         * See comments in g_typelib_symbol */
-        app_module = g_module_open (NULL, G_MODULE_BIND_LAZY);
-        if (app_module)
-          typelib->modules = g_list_append (typelib->modules, app_module);
+          if (module == NULL)
+            {
+              g_warning ("Failed to load shared library '%s' referenced by the typelib: %s",
+                         shlibs[i], g_module_error ());
+            }
+          else
+            {
+              typelib->modules = g_list_append (typelib->modules, module);
+            }
+       }
+
+      g_strfreev (shlibs);
     }
+  else
+    {
+      /* If there's no shared-library entry for this module, assume that
+       * the module is for the application.  Some of the hand-written .gir files
+       * in gobject-introspection don't have shared-library entries, but no one
+       * is really going to be calling g_module_symbol on them either.
+       */
+      GModule *module = g_module_open (NULL, 0);
+      if (module == NULL)
+        g_warning ("gtypelib.c: Failed to g_module_open (NULL): %s", g_module_error ());
+      else
+        typelib->modules = g_list_prepend (typelib->modules, module);
+    }
+}
+
+static inline void
+_g_typelib_ensure_open (GTypelib *typelib)
+{
+  if (typelib->open_attempted)
+    return;
+  typelib->open_attempted = TRUE;
+  _g_typelib_do_dlopen (typelib);
 }
 
 /**
@@ -2015,7 +2062,7 @@ g_typelib_new_from_memory (guchar *memory, gsize len)
   meta->len = len;
   meta->owns_memory = TRUE;
   meta->modules = NULL;
-  _g_typelib_init (meta);
+
   return meta;
 }
 
@@ -2038,7 +2085,7 @@ g_typelib_new_from_const_memory (const guchar *memory, gsize len)
   meta->len = len;
   meta->owns_memory = FALSE;
   meta->modules = NULL;
-  _g_typelib_init (meta);
+
   return meta;
 }
 
@@ -2060,7 +2107,7 @@ g_typelib_new_from_mapped_file (GMappedFile *mfile)
   meta->owns_memory = FALSE;
   meta->data = (guchar *) g_mapped_file_get_contents (mfile);
   meta->len = g_mapped_file_get_length (mfile);
-  _g_typelib_init (meta);
+
   return meta;
 }
 
@@ -2094,6 +2141,7 @@ g_typelib_get_namespace (GTypelib *typelib)
 
 /**
  * g_typelib_symbol:
+ * @typelib: the typelib
  * @symbol_name: name of symbol to be loaded
  * @symbol: returns a pointer to the symbol value
  *
@@ -2105,21 +2153,19 @@ gboolean
 g_typelib_symbol (GTypelib *typelib, const char *symbol_name, gpointer *symbol)
 {
   GList *l;
+  
+  _g_typelib_ensure_open (typelib);
 
   /*
-   * We want to be able to add symbols to an app or an auxiliary
-   * library to fill in gaps in an introspected library. However,
-   * normally we would only look for symbols in the main library
-   * (the first items in typelib->modules).
-   *
-   * A more elaborate solution is probably possible, but as a
-   * simple approach for now, if we fail to find a symbol we look
-   * for it in the global module (the last item in type->modules).
-   *
-   * This would not be very efficient if it happened often, since
-   * we always do the failed lookup above first, but very few
-   * symbols should be outside of the main libraries in
-   * typelib->modules so it doesn't matter.
+   * The reason for having multiple modules dates from gir-repository
+   * when it was desired to inject code (accessors, etc.) into an
+   * existing library.  In that situation, the first module listed
+   * will be the custom one, which overrides the main one.  A bit
+   * inefficient, but the problem will go away when gir-repository
+   * does.
+   * 
+   * For modules with no shared library, we dlopen'd the current
+   * process above.  
    */
   for (l = typelib->modules; l; l = l->next)
     {
