@@ -25,14 +25,15 @@ import tempfile
 import shutil
 import subprocess
 
-from .ast import (Callback, Constant, Enum, Function, Member, Namespace,
-                  Parameter, Property, Return, Struct, Type, Alias,
-                  Union, Field, type_name_from_ctype,
-                  default_array_types, TYPE_UINT8, PARAM_DIRECTION_IN)
+from .ast import (Alias, Bitfield, Callback, Constant, Enum, Function, Member,
+                  Namespace, Parameter, Property, Record, Return, Type, Union,
+                  Field, VFunction, type_name_from_ctype,
+                  default_array_types, TYPE_UINT8, PARAM_TRANSFER_FULL)
 from .transformer import Names
 from .glibast import (GLibBoxed, GLibEnum, GLibEnumMember, GLibFlags,
                       GLibInterface, GLibObject, GLibSignal, GLibBoxedStruct,
-                      GLibBoxedUnion, GLibBoxedOther, type_names)
+                      GLibBoxedUnion, GLibBoxedOther, GLibRecord,
+                      type_names)
 from .utils import to_underscores, to_underscores_noprefix
 
 default_array_types['guchar*'] = TYPE_UINT8
@@ -59,6 +60,18 @@ SYMBOL_BLACKLIST = [
 SYMBOL_BLACKLIST_RE = [re.compile(x) for x in \
                            [r'\w+_marshal_[A-Z]+__', ]]
 
+GET_TYPE_OVERRIDES = {
+    # this is a special case, from glibtransforer.py:create_gobject
+    'intern': 'g_object_get_type',
+    # this is presumably a typo, should be fixed upstream
+    'g_gstring_get_type': 'g_string_get_type',
+    # this is historical cruft: there's a deprecated
+    #   #define gdk_window_get_type gdk_window_get_window_type
+    # upstream; this method can be renamed properly upstream once
+    # that deprecated alias is removed (in some future release)
+    'gdk_window_object_get_type': 'gdk_window_get_type',
+}
+
 
 class IntrospectionBinary(object):
 
@@ -85,13 +98,12 @@ class GLibTransformer(object):
     def __init__(self, transformer, noclosure=False):
         self._transformer = transformer
         self._noclosure = noclosure
-        self._transformer.set_container_types(['GList*', 'GSList*'],
-                                              ['GHashTable*'])
         self._namespace_name = None
         self._names = Names()
         self._uscore_type_names = {}
         self._binary = None
         self._get_type_functions = []
+        self._error_quark_functions = []
         self._gtype_data = {}
         self._failed_types = {}
         self._boxed_types = {}
@@ -100,9 +112,6 @@ class GLibTransformer(object):
 
     # Public API
 
-    def set_introspection_binary(self, binary):
-        self._binary = binary
-
     def _print_statistics(self):
         nodes = list(self._names.names.itervalues())
 
@@ -115,9 +124,16 @@ class GLibTransformer(object):
         print " %d nodes; %d objects, %d interfaces, %d enums" \
             % (len(nodes), objectcount, ifacecount, enumcount)
 
-    def parse(self):
+    def init_parse(self):
+        """Do parsing steps that don't involve the introspection binary
+
+        This does enough work that get_type_functions() can be called.
+
+        """
+
         namespace = self._transformer.parse()
         self._namespace_name = namespace.name
+        self._namespace_version = namespace.version
 
         # First pass: parsing
         for node in namespace.nodes:
@@ -128,6 +144,15 @@ class GLibTransformer(object):
         if namespace.name == 'GObject':
             del self._names.aliases['Type']
 
+    def get_get_type_functions(self):
+        return self._get_type_functions
+
+    def set_introspection_binary(self, binary):
+        self._binary = binary
+
+    def parse(self):
+        """Do remaining parsing steps requiring introspection binary"""
+
         # Get all the GObject data by passing our list of get_type
         # functions to the compiled binary
 
@@ -145,21 +170,26 @@ class GLibTransformer(object):
             try:
                 self._resolve_node(node)
             except KeyError, e:
-                print "WARNING: DELETING node %s: %s" % (node.name, e)
+                #print "WARNING: DELETING node %s: %s" % (node.name, e)
                 self._remove_attribute(node.name)
+        # Another pass, since we need to have the methods parsed
+        # in order to correctly modify them after class/record
+        # pairing
+        for (ns, node) in nodes:
             # associate GtkButtonClass with GtkButton
-            if isinstance(node, Struct):
-                self._pair_class_struct(node)
+            if isinstance(node, Record):
+                self._pair_class_record(node)
         for (ns, alias) in self._names.aliases.itervalues():
             self._resolve_alias(alias)
-
-        self._print_statistics()
+        self._resolve_quarks()
         # Fourth pass: ensure all types are known
         if not self._noclosure:
-            self._validate(nodes)
+            self._resolve_types(nodes)
+
+        #self._validate(nodes)
 
         # Create a new namespace with what we found
-        namespace = Namespace(namespace.name, namespace.version)
+        namespace = Namespace(self._namespace_name, self._namespace_version)
         namespace.nodes = map(lambda x: x[1], self._names.aliases.itervalues())
         for (ns, x) in self._names.names.itervalues():
             namespace.nodes.append(x)
@@ -195,6 +225,10 @@ class GLibTransformer(object):
     def _register_internal_type(self, type_name, node):
         self._names.type_names[type_name] = (None, node)
         uscored = to_underscores(type_name).lower()
+        # prefer the prefix of the get_type method, if there is one
+        if hasattr(node, 'get_type'):
+            uscored = GET_TYPE_OVERRIDES.get(node.get_type, node.get_type)
+            uscored = uscored[:-len('_get_type')]
         self._uscore_type_names[uscored] = node
         # Besides the straight underscore conversion, we also try
         # removing the underscores from the namespace as a possible C
@@ -202,7 +236,16 @@ class GLibTransformer(object):
         suffix = self._transformer.remove_prefix(type_name)
         prefix = type_name[:-len(suffix)]
         no_uscore_prefixed = (prefix + '_' + to_underscores(suffix)).lower()
-        self._uscore_type_names[no_uscore_prefixed] = node
+        # since this is a guess, don't overwrite any 'real' prefix
+        if no_uscore_prefixed not in self._uscore_type_names:
+            self._uscore_type_names[no_uscore_prefixed] = node
+
+    def _resolve_quarks(self):
+        for node in self._error_quark_functions:
+            short = node.symbol[:-len('_quark')]
+            enum = self._uscore_type_names.get(short)
+            if enum is not None:
+                enum.error_quark = node.symbol
 
     # Helper functions
 
@@ -216,6 +259,7 @@ class GLibTransformer(object):
     def _execute_binary(self):
         in_path = os.path.join(self._binary.tmpdir, 'types.txt')
         f = open(in_path, 'w')
+        # TODO: Introspect GQuark functions
         for func in self._get_type_functions:
             f.write(func)
             f.write('\n')
@@ -227,7 +271,10 @@ class GLibTransformer(object):
         args.append('--introspect-dump=%s,%s' % (in_path, out_path))
 
         # Invoke the binary, having written our get_type functions to types.txt
-        subprocess.check_call(args, stdout=sys.stdout, stderr=sys.stderr)
+        try:
+            subprocess.check_call(args, stdout=sys.stdout, stderr=sys.stderr)
+        except subprocess.CalledProcessError, e:
+            raise SystemExit(e)
         self._read_introspect_dump(out_path)
 
         # Clean up temporaries
@@ -251,19 +298,34 @@ class GLibTransformer(object):
             parent_type_name = 'GObject'
             parent_gitype = self._resolve_gtypename(parent_type_name)
             symbol = 'g_initially_unowned_get_type'
-        node = GLibObject(node.name, parent_gitype, type_name, symbol, True)
-        self._add_attribute(node)
-        self._register_internal_type(type_name, node)
+        else:
+            assert False
+        gnode = GLibObject(node.name, parent_gitype, type_name, symbol, True)
+        if type_name == 'GObject':
+            gnode.fields.extend(node.fields)
+        else:
+            # http://bugzilla.gnome.org/show_bug.cgi?id=569408
+            # GInitiallyUnowned is actually a typedef for GObject, but
+            # that's not reflected in the GIR, where it appears as a
+            # subclass (as it appears in the GType hierarchy).  So
+            # what we do here is copy all of the GObject fields into
+            # GInitiallyUnowned so that struct offset computation
+            # works correctly.
+            gnode.fields = self._names.names['Object'][1].fields
+        self._add_attribute(gnode)
+        self._register_internal_type(type_name, gnode)
 
     # Parser
 
     def _parse_node(self, node):
         if isinstance(node, Enum):
             self._parse_enum(node)
+        elif isinstance(node, Bitfield):
+            self._parse_bitfield(node)
         elif isinstance(node, Function):
             self._parse_function(node)
-        elif isinstance(node, Struct):
-            self._parse_struct(node)
+        elif isinstance(node, Record):
+            self._parse_record(node)
         elif isinstance(node, Callback):
             self._parse_callback(node)
         elif isinstance(node, Alias):
@@ -284,6 +346,9 @@ class GLibTransformer(object):
     def _parse_enum(self, enum):
         self._add_attribute(enum)
 
+    def _parse_bitfield(self, enum):
+        self._add_attribute(enum)
+
     def _parse_constant(self, constant):
         self._add_attribute(constant)
 
@@ -297,6 +362,8 @@ class GLibTransformer(object):
                 return
         if self._parse_get_type_function(func):
             return
+        if self._parse_error_quark_function(func):
+            return
 
         self._add_attribute(func)
 
@@ -307,6 +374,10 @@ class GLibTransformer(object):
         if self._namespace_name == 'GLib':
             # No GObjects in GLib
             return False
+        if (self._namespace_name == 'GObject' and
+            symbol in ('g_object_get_type', 'g_initially_unowned_get_type')):
+            # We handle these internally, see _create_gobject
+            return True
         if func.parameters:
             return False
         # GType *_get_type(void)
@@ -321,6 +392,18 @@ class GLibTransformer(object):
         self._get_type_functions.append(symbol)
         return True
 
+    def _parse_error_quark_function(self, func):
+        if not func.symbol.endswith('_error_quark'):
+            return False
+        if func.parameters:
+            return False
+        if func.retval.type.name not in ['GLib.Quark',
+                                         'GQuark']:
+            return False
+
+        self._error_quark_functions.append(func)
+        return True
+
     def _name_is_internal_gtype(self, giname):
         try:
             node = self._get_attribute(giname)
@@ -336,7 +419,7 @@ class GLibTransformer(object):
         target_klass = None
         prefix_components = None
         methname = None
-        for i in xrange(1, len(components)-1):
+        for i in xrange(1, len(components)):
             prefix_components = '_'.join(components[0:-i])
             methname = '_'.join(components[-i:])
             target_klass = self._uscore_type_names.get(prefix_components)
@@ -348,6 +431,7 @@ class GLibTransformer(object):
         self._remove_attribute(func.name)
         func.name = methname
         target_klass.static_methods.append(func)
+        func.is_method = True
         return func
 
     def _parse_method(self, func):
@@ -380,6 +464,13 @@ class GLibTransformer(object):
             argtype = target_arg.type.ctype.replace('*', '')
             name = self._transformer.remove_prefix(argtype)
             name_uscore = to_underscores_noprefix(name).lower()
+            # prefer the prefix of the _get_type method, if there is one
+            if argtype in self._names.type_names:
+                node = self._names.type_names[argtype][1]
+                if hasattr(node, 'get_type'):
+                    name_uscore = GET_TYPE_OVERRIDES.get(node.get_type,
+                                                         node.get_type)
+                    name_uscore = name_uscore[:-len('_get_type')]
             name_offset = func.symbol.find(name_uscore)
             if name_offset < 0:
                 return None
@@ -410,9 +501,15 @@ class GLibTransformer(object):
 
         # The _uscore_type_names member holds the plain GLibBoxed
         # object; we want to actually use the struct/record associated
-        if isinstance(klass, GLibBoxed):
-            name = self._transformer.remove_prefix(klass.type_name)
-            klass = self._get_attribute(name)
+        if isinstance(klass, (Record, Union)):
+            remove_prefix = klass.symbol
+        else:
+            remove_prefix = klass.type_name
+
+        name = self._transformer.remove_prefix(remove_prefix)
+        klass = self._get_attribute(name)
+        if klass is None:
+            return
 
         if not is_method:
             # Interfaces can't have constructors, punt to global scope
@@ -424,7 +521,7 @@ class GLibTransformer(object):
             # class from the prefix
             # But for now, ensure that constructor returns are always
             # the most concrete class
-            name = self._transformer.remove_prefix(klass.type_name)
+            name = self._transformer.remove_prefix(remove_prefix)
             func.retval.type = Type(name, func.retval.type.ctype)
 
         self._remove_attribute(func.name)
@@ -434,32 +531,35 @@ class GLibTransformer(object):
             # We don't need the "this" parameter
             del func.parameters[0]
             klass.methods.append(func)
+            func.is_method = True
         else:
             klass.constructors.append(func)
         return func
 
-    def _parse_struct(self, struct):
+    def _parse_record(self, record):
         # This is a hack, but GObject is a rather fundamental piece so.
         internal_names = ["Object", 'InitiallyUnowned']
         g_internal_names = ["G" + x for x in internal_names]
         if (self._namespace_name == 'GObject' and
-            struct.name in internal_names):
-            self._create_gobject(struct)
+            record.name in internal_names):
+            self._create_gobject(record)
             return
-        elif struct.name in g_internal_names:
+        elif record.name in g_internal_names:
             # Avoid duplicates
             return
-        node = self._names.names.get(struct.name)
+        node = self._names.names.get(record.name)
         if node is None:
-            self._add_attribute(struct, replace=True)
+            self._add_attribute(record, replace=True)
+            self._register_internal_type(record.symbol, record)
             return
         (ns, node) = node
-        node.fields = struct.fields[:]
+        node.fields = record.fields[:]
 
     def _parse_union(self, union):
         node = self._names.names.get(union.name)
         if node is None:
             self._add_attribute(union, replace=True)
+            self._register_internal_type(union.symbol, union)
             return
         (ns, node) = node
         node.fields = union.fields[:]
@@ -484,14 +584,20 @@ class GLibTransformer(object):
             return True
         return False
 
-    def _pair_class_struct(self, maybe_class):
+    def _pair_class_record(self, maybe_class):
         name = self._strip_class_suffix(maybe_class.name)
         if name == maybe_class.name:
             return
 
-        if self._arg_is_failed(maybe_class):
-            print "WARNING: deleting no-type %r" % (maybe_class.name, )
-            del self._names.names[maybe_class.name]
+        class_struct = maybe_class
+        if self._arg_is_failed(class_struct):
+            print "WARNING: deleting no-type %r" % (class_struct.name, )
+            del self._names.names[class_struct.name]
+            return
+
+        pair_class = self._get_attribute(name)
+        if (not pair_class or
+            not isinstance(pair_class, (GLibObject, GLibInterface))):
             return
 
         # Object class fields are assumed to be read-only
@@ -500,20 +606,47 @@ class GLibTransformer(object):
             if isinstance(field, Field):
                 field.writable = False
 
-        name = self._resolve_type_name(name)
-        resolved = self._transformer.remove_prefix(name)
-        pair_class = self._get_attribute(resolved)
-        if pair_class and isinstance(pair_class,
-                                     (GLibObject, GLibInterface)):
-            for field in maybe_class.fields[1:]:
-                pair_class.fields.append(field)
-            return
-        name = self._transformer.remove_prefix(maybe_class.name)
-        pair_class = self._get_attribute(name)
-        if pair_class and isinstance(pair_class,
-                                     (GLibObject, GLibInterface)):
+        # Loop through fields to determine which are virtual
+        # functions and which are signal slots by
+        # assuming everything that doesn't share a name
+        # with a known signal is a virtual slot.
+        for field in maybe_class.fields:
+            if not isinstance(field, Callback):
+                continue
+            # Check the first parameter is the object
+            if len(field.parameters) == 0:
+                continue
+            firstparam_type = field.parameters[0].type
+            if firstparam_type != pair_class:
+                continue
+            # Also double check we don't have a signal with this
+            # name.
+            matched_signal = False
+            for signal in pair_class.signals:
+                if signal.name.replace('-', '_') == field.name:
+                    matched_signal = True
+                    break
+            if matched_signal:
+                continue
+            vfunc = VFunction.from_callback(field)
+            pair_class.virtual_methods.append(vfunc)
 
-            del self._names.names[maybe_class.name]
+        # Take the set of virtual methods we found, and try
+        # to pair up with any matching methods using the
+        # name+signature.
+        for vfunc in pair_class.virtual_methods:
+            for method in pair_class.methods:
+                if (method.name != vfunc.name or
+                    method.retval != vfunc.retval or
+                    method.parameters != vfunc.parameters):
+                    continue
+                vfunc.invoker = method.name
+
+        gclass_struct = GLibRecord.from_record(class_struct)
+        self._remove_attribute(class_struct.name)
+        self._add_attribute(gclass_struct, True)
+        pair_class.glib_type_struct = gclass_struct
+        gclass_struct.is_gtype_struct_for = name
 
     # Introspection
 
@@ -532,7 +665,10 @@ class GLibTransformer(object):
     def _introspect_enum(self, node):
         members = []
         for member in node.findall('member'):
-            members.append(GLibEnumMember(member.attrib['nick'],
+            # Keep the name closer to what we'd take from C by default;
+            # see http://bugzilla.gnome.org/show_bug.cgi?id=575613
+            name = member.attrib['nick'].replace('-', '_')
+            members.append(GLibEnumMember(name,
                                           member.attrib['value'],
                                           member.attrib['name'],
                                           member.attrib['nick']))
@@ -563,14 +699,14 @@ class GLibTransformer(object):
         self._introspect_signals(node, xmlnode)
         self._introspect_implemented_interfaces(node, xmlnode)
 
-        # add struct fields
-        struct = self._get_attribute(node.name)
-        if struct is not None:
-            node.fields = struct.fields
+        # add record fields
+        record = self._get_attribute(node.name)
+        if record is not None:
+            node.fields = record.fields
             for field in node.fields:
                 if isinstance(field, Field):
                     # Object instance fields are assumed to be read-only
-                    # (see also _pair_class_struct and transformer.py)
+                    # (see also _pair_class_record and transformer.py)
                     field.writable = False
 
         self._add_attribute(node, replace=True)
@@ -632,7 +768,7 @@ class GLibTransformer(object):
             rctype = signal_info.attrib['return']
             rtype = Type(self._transformer.parse_ctype(rctype), rctype)
             return_ = Return(rtype, signal_info.attrib['return'])
-            return_.transfer = 'full'
+            return_.transfer = PARAM_TRANSFER_FULL
             signal = GLibSignal(signal_info.attrib['name'], return_)
             for i, parameter in enumerate(signal_info.findall('param')):
                 if i == 0:
@@ -676,8 +812,8 @@ class GLibTransformer(object):
             self._resolve_glib_object(node)
         elif isinstance(node, GLibInterface):
             self._resolve_glib_interface(node)
-        elif isinstance(node, Struct):
-            self._resolve_struct(node)
+        elif isinstance(node, Record):
+            self._resolve_record(node)
         elif isinstance(node, Union):
             self._resolve_union(node)
         elif isinstance(node, Alias):
@@ -699,7 +835,7 @@ class GLibTransformer(object):
         if not pair_node:
             boxed_item = GLibBoxedOther(name, boxed.type_name,
                                         boxed.get_type)
-        elif isinstance(pair_node, Struct):
+        elif isinstance(pair_node, Record):
             boxed_item = GLibBoxedStruct(pair_node.name, boxed.type_name,
                                          boxed.get_type)
             boxed_item.fields = pair_node.fields
@@ -711,7 +847,7 @@ class GLibTransformer(object):
             return False
         self._add_attribute(boxed_item, replace=True)
 
-    def _resolve_struct(self, node):
+    def _resolve_record(self, node):
         for field in node.fields:
             self._resolve_field(field)
 
@@ -747,7 +883,17 @@ class GLibTransformer(object):
              for x in node.prerequisites])
 
     def _resolve_glib_object(self, node):
-        node.parent = self._force_resolve(node.parent)
+        # If we can't find the parent class, just drop back to GObject.
+        # This supports hidden parent classes.
+        # http://bugzilla.gnome.org/show_bug.cgi?id=561360
+        try:
+            node.parent = self._force_resolve(node.parent)
+        except KeyError, e:
+            #print ("WARNING: Parent %r of class %r" +\
+            #       " not found; using GObject") % (node.parent.target,
+            #                                       node.name)
+            node.parent = self._transformer.gtypename_to_giname("GObject",
+                                                                self._names)
         node.interfaces = filter(None,
             [self._force_resolve(x, allow_unknown=True)
                                     for x in node.interfaces])
@@ -783,32 +929,13 @@ class GLibTransformer(object):
             except KeyError, e:
                 failed.append(prop)
         for fail in failed:
-            print ("WARNING: Deleting object property %r (of %r) "
-                   "with unknown type") % (fail, context)
+            #print ("WARNING: Deleting object property %r (of %r) "
+            #       "with unknown type") % (fail, context)
             properties.remove(fail)
 
     def _resolve_property(self, prop):
         prop.type = self._resolve_param_type(prop.type, allow_invalid=False)
 
-    def _adjust_transfer(self, param):
-        if not (param.transfer is None or param.transfer_inferred):
-            return
-
-        # Do GLib/GObject-specific type transformations here
-        node = self._lookup_node(param.type.name)
-        if node is None:
-            return
-
-        if isinstance(param, Parameter):
-            if param.direction != PARAM_DIRECTION_IN:
-                transfer = 'full'
-            else:
-                transfer = 'none'
-        else:
-            transfer = 'full'
-
-        param.transfer = transfer
-
     def _adjust_throws(self, func):
         if func.parameters == []:
             return
@@ -826,32 +953,32 @@ class GLibTransformer(object):
         self._resolve_parameters(func.parameters)
         func.retval.type = self._resolve_param_type(func.retval.type)
         self._adjust_throws(func)
-        self._adjust_transfer(func.retval)
 
     def _resolve_parameters(self, parameters):
         for parameter in parameters:
             parameter.type = self._resolve_param_type(parameter.type)
-            self._adjust_transfer(parameter)
 
     def _resolve_field(self, field):
         if isinstance(field, Callback):
             self._resolve_function(field)
-            return
-        field.type = self._resolve_param_type(field.type)
+        elif isinstance(field, Record): # non-typedef'd struct
+            self._resolve_record(field)
+        elif isinstance(field, Union): # non-typedef'd union
+            self._resolve_union(field)
+        else:
+            field.type = self._resolve_param_type(field.type)
 
     def _resolve_alias(self, alias):
         alias.target = self._resolve_type_name(alias.target, alias.target)
 
-    # Validation
-
-    def _validate(self, nodes):
+    def _resolve_types(self, nodes):
         nodes = list(self._names.names.itervalues())
         i = 0
         self._validating = True
         while True:
             initlen = len(nodes)
 
-            print "Type resolution; pass=%d" % (i, )
+            #print "Type resolution; pass=%d" % (i, )
             nodes = list(self._names.names.itervalues())
             for node in nodes:
                 try:
@@ -864,3 +991,18 @@ class GLibTransformer(object):
             i += 1
             self._print_statistics()
         self._validating = False
+
+    # Validation
+
+    def _validate_interface(self, iface):
+        for vfunc in iface.virtual_methods:
+            if not vfunc.invoker:
+                print ("warning: Interface %r virtual function %r " + \
+                       "has no known invoker") % (iface.name, vfunc.name)
+
+    # This function is called at the very end, before we hand back the
+    # completed namespace to the writer.  Add static analysis checks here.
+    def _validate(self, nodes):
+        for (name, node) in nodes:
+            if isinstance(node, GLibInterface):
+                self._validate_interface(node)
