@@ -4,6 +4,12 @@
 #include <glib-object.h>
 #include "utility.h"
 
+#define FOO_SUCCESS_INT 0x1138
+
+#define FOO_DEFINE_SHOULD_BE_EXPOSED "should be exposed"
+
+#define FOO_PIE_IS_TASTY 3.14159
+
 #define FOO_TYPE_INTERFACE           (foo_interface_get_type ())
 #define FOO_INTERFACE(object)        (G_TYPE_CHECK_INSTANCE_CAST ((object), FOO_TYPE_INTERFACE, FooInterface))
 #define FOO_IS_INTERFACE(object)     (G_TYPE_CHECK_INSTANCE_TYPE ((object), FOO_TYPE_INTERFACE))
@@ -23,6 +29,10 @@
 #define FOO_SUBOBJECT(subobject)     (G_TYPE_CHECK_INSTANCE_CAST ((subobject), FOO_TYPE_SUBOBJECT, FooSubobject))
 #define FOO_IS_SUBOBJECT(subobject)  (G_TYPE_CHECK_INSTANCE_TYPE ((subobject), FOO_TYPE_SUBOBJECT))
 
+#define FOO_TYPE_BUFFER        (foo_buffer_get_type ())
+#define FOO_BUFFER(object)     (G_TYPE_CHECK_INSTANCE_CAST ((object), FOO_TYPE_BUFFER, FooBuffer))
+#define FOO_IS_BUFFER(object)  (G_TYPE_CHECK_INSTANCE_TYPE ((object), FOO_TYPE_BUFFER))
+
 typedef struct _FooInterface       FooInterface;
 typedef struct _FooInterfaceIface  FooInterfaceIface;
 typedef struct _FooSubInterface       FooSubInterface;
@@ -31,20 +41,30 @@ typedef struct _FooObject          FooObject;
 typedef struct _FooObjectClass     FooObjectClass;
 typedef struct _FooSubobject       FooSubobject;
 typedef struct _FooSubobjectClass  FooSubobjectClass;
+typedef struct _FooBuffer          FooBuffer;
+typedef struct _FooBufferClass     FooBufferClass;
 
 struct _FooInterfaceIface
 {
   GTypeInterface parent_iface;
 
-  void (*do_foo) (FooInterface *self);
+  void (*do_foo) (FooInterface *self, int x);
 };
 
 GType                 foo_interface_get_type       (void) G_GNUC_CONST;
 
+void foo_interface_do_foo (FooInterface *iface, int x);
+
 struct _FooSubInterfaceIface
 {
   GTypeInterface parent_iface;
 
+  /* signals */
+
+  void (*destroy_event) (FooSubInterface *self);
+
+  /* virtual table */
+
   void (*do_bar) (FooSubInterface *self);
 };
 
@@ -55,6 +75,8 @@ void foo_sub_interface_do_bar (FooSubInterface *self);
 struct _FooObject
 {
   GObject parent_instance;
+
+  int some_int;
 };
 
 struct _FooObjectClass
@@ -62,6 +84,9 @@ struct _FooObjectClass
   GObjectClass parent_class;
 
   gboolean (* virtual_method) (FooObject *object, int first_param);
+
+  /* Intended to match GFile */
+  void (*read_fn) (FooObject *object, int offset, int length);
 };
 
 gint                  foo_init                     (void);
@@ -91,7 +116,11 @@ char *                foo_object_dup_name          (FooObject *object);
 
 void                  foo_object_handle_glyph      (FooObject *object, UtilityGlyph glyph);
 
-int                   foo_object_static_meth       ();
+gboolean              foo_object_virtual_method    (FooObject *object, int first_param);
+
+void                  foo_object_read              (FooObject *object, int offset, int length);
+
+int                   foo_object_static_meth       (void);
 
 struct _FooSubobject
 {
@@ -104,7 +133,13 @@ struct _FooSubobjectClass
 };
 
 GType                 foo_subobject_get_type       (void) G_GNUC_CONST;
-FooSubobject*         foo_subobject_new            ();
+FooSubobject*         foo_subobject_new            (void);
+
+FooObject *           foo_object_get_default       (void);
+
+GType                 foo_buffer_get_type          (void);
+
+void                  foo_buffer_some_method       (FooBuffer *buffer);
 
 typedef enum
 {
@@ -126,6 +161,8 @@ typedef enum
   FOO_FLAGS_THIRD  = 1 << 2
 } FooFlagsType;
 
+GType foo_flags_type_get_type (void);
+
 typedef enum
 {
   FOO_ENUM_UN = 1,
@@ -134,7 +171,12 @@ typedef enum
   FOO_ENUM_NEUF = 9
 } FooEnumNoType;
 
-GType foo_flags_type_get_type (void);
+typedef enum
+{
+  FOO_FLAGS_ETT  = 1 << 0,
+  FOO_FLAGS_TVA = 1 << 1,
+  FOO_FLAGS_FYRA  = 1 << 2
+} FooFlagsNoType;
 
 typedef enum
 {
@@ -279,4 +321,53 @@ void foo_test_string_array (char **array);
 
 void foo_test_string_array_with_g (gchar **array);
 
+const char * foo_test_const_char_retval (void);
+const FooStruct * foo_test_const_struct_retval (void);
+void foo_test_const_char_param (const char * param);
+void foo_test_const_struct_param (const FooStruct * param);
+
+typedef void (*FooVarargsCallback) (const char * param, ...);
+void foo_test_varargs_callback (gint i, FooVarargsCallback callback);
+void foo_test_varargs_callback2 (FooVarargsCallback callback);
+void foo_test_varargs_callback3 (FooVarargsCallback callback,
+				 FooVarargsCallback callback2);
+
+typedef enum {
+  FOO_ERROR_GOOD,
+  FOO_ERROR_BAD,
+  FOO_ERROR_UGLY
+} FooError;
+GType foo_error_get_type (void);
+
+GQuark foo_error_quark (void);
+
+typedef enum
+{
+  FOO_LAYER_DESKTOP        = 0,
+  FOO_LAYER_BOTTOM        = 1,
+  FOO_LAYER_NORMAL        = 2,
+  FOO_LAYER_TOP        = 4, /* Same as DOCK; see EWMH and bug 330717 */
+  FOO_LAYER_DOCK        = 4,
+  FOO_LAYER_FULLSCREEN        = 5,
+  FOO_LAYER_FOCUSED_WINDOW    = 6,
+  FOO_LAYER_OVERRIDE_REDIRECT = 7,
+  FOO_LAYER_LAST        = 8
+} FooStackLayer;
+
+typedef enum
+{
+  FOO_SOME_SINGLE_ENUM
+} FooASingle;
+
+/* Should be skipped */
+void foo_some_variant (guint x, va_list args);
+void foo_some_variant_ptr (guint x, va_list *args);
+
+/* Should be skipped due to annotations */
+typedef enum {
+  FOO_SKIPPABLE_ONE,
+  FOO_SKIPPABLE_TWO
+} FooSkippable;
+void foo_skip_me (FooSkippable fs);
+
 #endif /* __FOO_OBJECT_H__ */
